@@ -28,7 +28,6 @@ __all__ = [
     "pd",
     "pdfilter",
     "pdb",
-    "plt",
     "PIL",
     "print",
     "puttext",
@@ -41,7 +40,6 @@ __all__ = [
     "resize",
     "rotate",
     "see",
-    "set_logging_level",
     "show",
     "store_attr",
     "subplots",
@@ -59,61 +57,89 @@ __all__ = [
     "xywh2xyXY",
     "df2bbs",
     "bbs2df",
+    # ---- Logging funcs ---- #
     "Info",
     "Warn",
     "Debug",
     "Excep",
-    "reset_logger_width",
-    "display",
+    "reset_logger",
+    "get_logger_level",
+    "in_debug_mode",
+    "debug_mode",
+    # "display",
     "typedispatch",
     "defaultdict",
     "Counter",
     "dcopy",
+    "patch_to",
+    "split",
+    "train_test_split",
+    "init_plt",
+    "init_cv2",
 ]
 
-
-import lovely_tensors as lt
-
-lt.monkey_patch()
-
-from .logger import *
-from .bb_utils import *
+import os
+import re
+import sys
+from builtins import print
 from pathlib import Path
-from fastcore.foundation import L
-from fastcore.dispatch import typedispatch
 
-import glob, numpy as np, pandas as pd, tqdm, os, sys, re
-from IPython.display import display
+import numpy as np
+import pandas as pd
 import PIL
+import tqdm
+from fastcore.all import L, delegates, patch_to
+from fastcore.dispatch import typedispatch
 from PIL import Image
 
-try:
-    import torch
-    import torch.nn as nn
-    from torch import optim
-    from torch.nn import functional as F
-    from torch.utils.data import Dataset, DataLoader
+# from .bb_utils import *
+# from .logger import *
 
-    __all__ += ["torch", "nn", "F", "Dataset", "DataLoader", "optim"]
-except:
-    ...
-import matplotlib  # ; matplotlib.use('Agg')
-import matplotlib.pyplot as plt
-import matplotlib.patheffects as path_effects
-import pdb, datetime
-from typing import Union, Tuple
+from .bb_utils import (
+    randint,
+    BB,
+    df2bbs,
+    bbs2df,
+    bbfy,
+    jitter,
+    enlarge_bbs,
+    shrink_bbs,
+    to_relative,
+    to_absolute,
+)
+from .logger import (
+    logger,
+    Info,
+    Warn,
+    Debug,
+    Excep,
+    reset_logger,
+    get_logger_level,
+    in_debug_mode,
+    debug_mode,
+)
 
+import datetime
+import pdb
+from typing import Tuple, Union
+
+# Aliases
 E = enumerate
+pd.read_pqt = pd.read_parquet
 
-try:
-    import cv2
+from fastcore.foundation import coll_repr, is_array
 
-    __all__ += ["cv2"]
-except:
-    logger.warning("Skipping cv2 import")
 
-import time
-from collections import defaultdict, Counter
+@patch_to(L)
+def _repr_pretty_(self, p, cycle):
+    p.text(
+        "..."
+        if cycle
+        else repr(self.items) if is_array(self.items) else coll_repr(self, 20)
+    )
+
+
+from collections import Counter, defaultdict
 from copy import deepcopy as dcopy
 
 from rich.progress import track as _track
@@ -153,11 +179,14 @@ nunique = lambda l: len(set(l))
 
 
 @typedispatch
-def choose(List, n=1):
+def choose(List, n=1, verbose=True):
     if n == 1:
-        return List[randint(len(List))]
+        o = List[randint(len(List))]
     else:
-        return L([choose(List) for _ in range(n)])
+        o = L([choose(List) for _ in range(n)])
+    if verbose:
+        Info(f"Chose `{o}` from input")
+    return o
 
 
 @typedispatch
@@ -181,7 +210,7 @@ def choose(i: pd.DataFrame, n=1):
 
 
 rand = lambda n=6: "".join(
-    choose(list("1234567890qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM"), n=n)
+    choose(list("1234567890qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM"), n=n, verbose=False)
 )
 
 
@@ -196,10 +225,12 @@ def Tqdm(x, total=None, desc=None):
 
 from tqdm import trange
 
-now = lambda: str(datetime.datetime.now())[:-10].replace(" ", "_")
+old_now = lambda: str(datetime.datetime.now())[:-10].replace(" ", "_")
+now = lambda: f"{datetime.datetime.now():%Y%m%d-%H%M}"
 
 
-def read(fname, mode=0):
+def read(fname, mode=1):
+    init_cv2()
     img = cv2.imread(str(fname), mode)
     if mode == 1:
         img = img[..., ::-1]  # BGR to RGB
@@ -212,17 +243,23 @@ def readPIL(fname, mode="RGB"):
     return Image.open(str(fname)).convert(mode.upper())
 
 
-def crop_from_bb(im, bb):
+def crop_from_bb(im, bb, padding=None):
     if isinstance(bb, list):
-        return [crop_from_bb(im, _bb) for _bb in bb]
+        return [crop_from_bb(im, _bb, padding=padding) for _bb in bb]
     x, y, X, Y = bb
+    px, py, pX, pY = padding
     if max(x, y, X, Y) < 1.5:
         h, w = im.shape[:2]
         x, y, X, Y = BB(bb).absolute((h, w))
+    y = max(0, y - py)
+    Y = min(h, Y + pY)
+    x = max(0, x - px)
+    X = min(w, X + pX)
     return im.copy()[y:Y, x:X]
 
 
 def rect(im, bb, c=None, th=2):
+    init_cv2()
     c = "g" if c is None else c
     _d = {"r": (255, 0, 0), "g": (0, 255, 0), "b": (0, 0, 255), "y": (255, 0, 255)}
     c = _d[c] if isinstance(c, str) else c
@@ -243,17 +280,9 @@ def C(im):
         return np.repeat(im[..., None], 3, 2)
 
 
-def common_old(a, b):
-    """Wrapper around set intersection"""
-    x = set(a).intersection(set(b))
-    logger.opt(depth=1).log(
-        "INFO",
-        f"{len(x)} items found common from containers of {len(a)} and {len(b)} items respectively",
-    )
-    return set(sorted(x))
-
-
 def common(*items, silent=True):
+    from .logger import logger
+
     """Wrapper around set intersection"""
     x = set(items[0])
     for item in items[1:]:
@@ -268,6 +297,8 @@ def common(*items, silent=True):
 
 
 def diff(a, b, rev=False, silent=False):
+    from .logger import logger
+
     if not rev:
         o = set(sorted(set(a) - set(b)))
     else:
@@ -278,6 +309,7 @@ def diff(a, b, rev=False, silent=False):
 
 
 def puttext(im, string, org, scale=1, color=(255, 0, 0), thickness=2):
+    init_cv2()
     x, y = org
     org = x, int(y + 30 * scale)
     cv2.putText(im, str(string), org, cv2.FONT_HERSHEY_COMPLEX, scale, color, thickness)
@@ -296,6 +328,34 @@ def _jitter(i):
     return i + np.random.randint(4)
 
 
+def is_in_notebook():
+    try:
+        import importlib
+
+        # Test adapted from tqdm.autonotebook: https://github.com/tqdm/tqdm/blob/master/tqdm/autonotebook.py
+        get_ipython = sys.modules["IPython"].get_ipython
+        if "IPKernelApp" not in get_ipython().config:
+            raise ImportError("console")
+
+        return importlib.util.find_spec("IPython") is not None
+    except (AttributeError, ImportError, KeyError):
+        return False
+
+
+def init_plt():
+    import matplotlib.patheffects as path_effects
+    import matplotlib.pyplot as plt
+
+    plt.rcParams["axes.edgecolor"] = "black"
+    globals().update(locals())
+
+
+def init_cv2():
+    import cv2
+
+    globals().update(locals())
+
+
 def show(
     img=None,
     ax=None,
@@ -306,26 +366,67 @@ def show(
     texts=None,
     bb_colors=None,
     cmap="gray",
-    grid=False,
-    save_path=None,
-    text_sz=None,
-    df=None,
+    grid: bool = False,
+    save_path: str = None,
+    text_sz: int = None,
+    df: pd.DataFrame = None,
     pts=None,
     conns=None,
+    interactive: bool = False,
+    jitter: int = None,
+    frame_count: int = 1,
     **kwargs,
 ):
     "show an image"
+    from IPython.display import display, display_html
+
+    init_plt()
 
     try:
         if isinstance(img, (str, Path)):
             img = read(str(img), 1)
-        if isinstance(img, torch.Tensor):
-            img = img.cpu().detach().numpy().copy()
+        try:
+            import torch
+
+            if isinstance(img, torch.Tensor):
+                img = img.cpu().detach().numpy().copy()
+        except ModuleNotFoundError:
+            pass
         if isinstance(img, PIL.Image.Image):
             img = np.array(img)
-
     except Exception as e:
-        print(e)
+        Warn(e)
+
+    if title is None:
+        import inspect as I
+
+        frame = I.currentframe()
+        for _ in range(frame_count):
+            frame = frame.f_back
+        for var_name, var_val in frame.f_locals.items():
+            if var_val is img and var_name != "_" and not var_name.strip("_").isdigit():
+                title = var_name
+
+    if isinstance(img, pd.DataFrame):
+        df = img
+        max_rows = kwargs.pop("max_rows", 30)
+        max_rows = 10000 if max_rows == -1 else max_rows
+        if is_in_notebook():
+            html_str = ""
+            html_str += '<th style="text-align:center"><td style="vertical-align:top">'
+            if title is not None:
+                html_str += f'<h3 style="text-align: center;">{title}</h3>'
+            html_str += (
+                df.to_html(max_rows=max_rows)
+                .replace("table", 'table style="display:inline"')
+                .replace(' style="display:inline"', "")
+            )
+            html_str += "</td></th>"
+            display_html(html_str, raw=True)
+        else:
+            o = df.to_markdown()
+            print(o)
+        return
     if not isinstance(img, np.ndarray):
         display(img)
         return
@@ -337,6 +438,13 @@ def show(
     if img.max() == 255:
         img = img.astype(np.uint8)
     h, w = img.shape[:2]
+
+    if interactive:
+        from .interactive_show import ishow
+
+        ishow(img, df=df)
+        return
+
     if sz is None:
         if w < 50:
             sz = 1
@@ -357,14 +465,17 @@ def show(
         _show = False
 
     if df is not None:
-        try:
-            text_col = kwargs.pop("text_col", "text")
-            if text_col == "ixs":
-                texts = df.index.tolist()
-            else:
-                texts = df[text_col]
-        except:
-            pass
+        if isinstance(df, (str, Path)):
+            df = str(df)
+            df = pd.read_csv(df) if df.endswith("csv") else pd.read_parquet(df)
+
+        text_col = kwargs.pop("text_col", "info" if "info" in df.columns else "text")
+        if text_col == "ixs":
+            texts = df.index.tolist()
+        elif text_col is not None and text_col in df.columns:
+            texts = df[text_col]
+        if "color" in df.columns:
+            bb_colors = df["color"].tolist()
         bbs = df2bbs(df)  # assumes df has 'x,y,X,Y' columns or a single 'bb' column
     kwargs.pop("text_col") if "text_col" in kwargs else ...
     if isinstance(texts, pd.core.series.Series):
@@ -389,10 +500,15 @@ def show(
             if isinstance(bbs, torch.Tensor):
                 bbs = bbs.cpu().detach().numpy()
             bbs = bbs.astype(np.uint32).tolist()
-        _x_ = np.array(bbs).max()
+        if len(bbs) > 0:
+            _x_ = np.array(bbs).max()
+        else:
+            raise ValueError("Trying to plot with 0 bounding boxes...")
         rel = True if _x_ < 1.5 else False
         if rel:
             bbs = [BB(bb).absolute((h, w)) for bb in bbs]
+        if jitter:
+            bbs = [bb.jitter(jitter) for bb in bbs]
         bb_colors = (
             [[randint(255) for _ in range(3)] for _ in range(len(bbs))]
             if bb_colors == "random"
@@ -403,7 +519,7 @@ def show(
         img = C(img) if len(img.shape) == 2 else img
         [rect(img, tuple(bb), c=bb_colors[ix], th=th) for ix, bb in enumerate(bbs)]
     text_sz = text_sz if text_sz else (max(sz) * 3 // 5)
-    if texts is not None:
+    if texts is not None or texts == "ixs":
         if hasattr(texts, "shape"):
             if isinstance(texts, torch.Tensor):
                 texts = texts.cpu().detach().numpy()
@@ -461,6 +577,7 @@ def show(
 
 
 def puttext(ax, string, org, size=15, color=(255, 0, 0), thickness=2):
+    init_plt()
     x, y = org
     va = "top" if y < 15 else "bottom"
     text = ax.text(x, y, str(string), color="red", ha="left", va=va, size=size)
@@ -468,7 +585,9 @@ def puttext(ax, string, org, size=15, color=(255, 0, 0), thickness=2):
         [path_effects.Stroke(linewidth=3, foreground="white"), path_effects.Normal()]
     )
 
+
 def subplots(ims, nc=5, figsize=(5, 5), silent=True, **kwargs):
+    init_plt()
     if len(ims) == 0:
         return
     titles = kwargs.pop("titles", [None] * len(ims))
@@ -487,7 +606,7 @@ def subplots(ims, nc=5, figsize=(5, 5), silent=True, **kwargs):
     figsize = kwargs.pop("sz", figsize)
     figsize = (figsize, figsize) if isinstance(figsize, int) else figsize
     fig, axes = plt.subplots(nr, nc, figsize=figsize)
-    return_axes = kwargs.pop('return_axes', False)
+    return_axes = kwargs.pop("return_axes", False)
     axes = axes.flat
     fig.suptitle(kwargs.pop("suptitle", ""))
     dfs = kwargs.pop("dfs", [None] * len(ims))
@@ -516,21 +635,13 @@ def subplots(ims, nc=5, figsize=(5, 5), silent=True, **kwargs):
     plt.show()
 
 
-class L_old(list):
-    def __getitem__(self, keys):
-        if isinstance(keys, (int, slice)):
-            return list.__getitem__(self, keys)
-        return L([self[k] for k in keys])
-
-    def sample(self, n=1):
-        return [self[randint(len(self))] for _ in range(n)]
-
-
 uint = lambda im: (255 * im).astype(np.uint8)
 Blank = lambda *sh: uint(np.ones(sh))
 
 
 def pdfilter(df, column, condition, silent=True):
+    from .logger import logger
+
     if not callable(condition):
         if isinstance(condition, list):
             condition = lambda x: x in condition
@@ -546,32 +657,6 @@ def pdsort(df, column, asc=True):
     df.sort_values(column, ascending=asc)
 
 
-def set_logging_level(level):
-    logger.remove()
-    logger.add(sys.stderr, level=level)
-
-
-def resize_old(im: np.ndarray, sz: Union[float, Tuple[int, int]]):
-    h, w = im.shape[:2]
-    if isinstance(sz, float):
-        frac = sz
-        H, W = [int(i * frac) for i in [h, w]]
-    elif isinstance(sz, int):
-        H, W = sz, sz
-    elif isinstance(sz, tuple):
-        if sz[0] == -1:
-            _, W = sz
-            f = W / w
-            H = int(f * h)
-        elif sz[1] == -1:
-            H, _ = sz
-            f = H / h
-            W = int(f * w)
-        else:
-            H, W = sz
-    return cv2.resize(im, (W, H))
-
-
 def resize(
     im: Union[np.ndarray, PIL.Image.Image],
     sz: Union[float, Tuple[int, int], Tuple[str, Tuple[int, int]]],
@@ -598,6 +683,7 @@ def resize(
         im = np.array(im)
         to_pil = True
     else:
+        init_cv2()
         to_pil = False
     h, w = im.shape[:2]
     if isinstance(sz, (tuple, list)) and isinstance(sz[0], str):
@@ -689,6 +775,7 @@ def parent(fpath):
 
 
 def write(image, fpath):
+    init_cv2()
     makedir(parent(fpath))
     cv2.imwrite(fpath, image)
 
@@ -697,49 +784,6 @@ def lzip(*x):
     return list(zip(*x))
 
 
-# def to_absolute(input, shape):
-#     if isinstance(shape, np.ndarray) and shape.ndim >= 2:
-#         shape = shape.shape[:2]
-#     h, w = shape
-#     if isinstance(input, BB):
-#         return input.absolute((h, w))
-#     elif isinstance(input, pd.DataFrame):
-
-#         def _round(x):
-#             return np.round(x.values.astype(np.double)).astype(np.uint16)
-
-#         df = input.copy()
-#         df["x"] = _round(df.x * w)
-#         df["X"] = _round(df.X * w)
-#         df["y"] = _round(df.y * h)
-#         df["Y"] = _round(df.Y * h)
-#         return df
-#     elif isinstance(input, list):
-#         bbs = bbfy(input)
-#         return [bb.absolute((h, w)) for bb in bbs]
-
-
-# def to_relative(input, shape):
-#     if isinstance(shape, np.ndarray) and shape.ndim >= 2:
-#         shape = shape.shape[:2]
-#     h, w = shape
-#     if isinstance(input, BB):
-#         return input.relative((h, w))
-#     elif isinstance(input, pd.DataFrame):
-#         df = input.copy()
-#         df["x"] = df.x / w
-#         df["X"] = df.X / w
-#         df["y"] = df.y / h
-#         df["Y"] = df.Y / h
-#         return df
-#     elif isinstance(input, list):
-#         bbs = bbfy(input)
-#         return [bb.relative((h, w)) for bb in bbs]
-
-
-from fastcore.basics import patch_to
-
-
 @patch_to(L)
 def first(self):
     if len(self) > 0:
@@ -763,16 +807,85 @@ def get_all(self, condition):
     return sublist
 
 
-def batchify(items, batch_size=32):
-    "yield `batch_size` items at a time"
-    head = 0
+def batchify(items, *rest, batch_size=32):
+    """
+    Yield batches of `batch_size` items at a time from one or more sequences.
+
+    Parameters:
+        items (iterable): The main sequence of items to be batchified.
+        *rest (iterable, optional): Additional sequences that should have the same length as `items`.
+        batch_size (int, optional): The number of items per batch.
+
+    Yields:
+        tuple: A tuple containing batches of items from `items` and optionally from the additional sequences.
+    """
     n = len(items)
+    if len(rest) > 0:
+        assert all(
+            [n == len(_rest) for _rest in rest]
+        ), "batchify can only work with equal length containers"
+    head = 0
     while head < n:
         tail = head + batch_size
         batch = items[head:tail]
-        yield batch
+        if rest:
+            rest_ = [list(_rest[head:tail]) for _rest in rest]
+            # yield batch, *rest_
+            yield None
+        else:
+            yield batch
         head = tail
 
 
 def toss(frac: float, pass_=True, fail_=False):
     return [fail_, pass_][np.random.uniform() < frac]
+
+
+def phasify(items, n_phases: int):
+    """
+    [Doc generated by AI]
+    Distributes a list of items into multiple phases and returns a list of lists, each containing items from a specific phase.
+
+    Parameters:
+        items (list): A list of items to be distributed among phases.
+        n_phases (int): The number of phases to divide the items into.
+
+    Returns:
+        list: A list of lists, where each sublist contains items assigned to a specific phase.
+
+    Example:
+        items = [1, 2, 3, 4, 5, 6, 7, 8, 9]
+        n_phases = 3
+        result = phasify(items, n_phases)
+        # Output: [[1, 4, 7], [2, 5, 8], [3, 6, 9]]
+    """
+    iterators = defaultdict(L)
+    [iterators[ix % n_phases].append(item) for ix, item in enumerate(items)]
+    return L(iterators.values())
+
+
+def split(items, splits, random_state=10):
+    ks, vs = lzip(*splits.items())
+    if any([v == -1 for v in vs]):
+        assert list(vs).count(-1) == 1, "Only atmost one `-1` is allowed"
+        vs = [v if v != -1 else -sum(vs) for v in vs]
+    Info(vs)
+    assert sum(vs) == 1, f"Split percentages should add to 1, received sum={sum(vs)}"
+    np.random.seed(random_state)
+    assignments = np.random.choice(ks, size=len(items), p=vs)
+    o = {k: [] for k in ks}
+    for item, assigned in zip(items, assignments):
+        o[assigned].append(item)
+    return o
+
+
+def train_test_split(*args, **kwargs):
+    # This is done mainly to save time and memory during imports
+    from sklearn.model_selection import train_test_split as tts
+
+    return tts(*args, **kwargs)
+
+
+@patch_to(L)
+def to_json(self):
+    return list(self)
